class MountaineerUnsetValue:
    """
    A sentinel value to indicate that a value has not been set. This is useful for cases where
    we need to separate explicit Nones from missing values.
    """

    pass
